/*
 * xgif.c - displays GIF pictures on an X11 display
 *
 *  Author:    John Bradley, University of Pennsylvania
 *                (bradley@cis.upenn.edu)
 */

#define MAIN
#include "xgif.h"

int wcount = 0;
Window assoc_w[10];
int assoc_theH[10];
int assoc_theW[10];
XImage *assoc_theImg[10];
XImage *assoc_expImg[10];

Window gif_CreateMainWindow();
XImage *gif_Resize();

int findAssocIndex(w)
	Window w;
{
	int i;
	for (i = 0; i < wcount; i++) if (assoc_w[i] == w) return i;
	return -1;
}

XImage *gif_load(fname, w, x, y, width, height, actual_w, actual_h, theImgp)
	char *fname;
	Window w;
	int x, y, width, height;
	int *actual_w, *actual_h;
	XImage **theImgp;
{
	XImage *expImg;

	theImage = NULL;
	gif_reset();
	LoadGIF(fname);
	*theImgp = theImage;

	iWIDE = theImage->width;  iHIGH = theImage->height;
	*actual_w = theImage->width; *actual_h = theImage->height;

	eWIDE = iWIDE * expand;  eHIGH = iHIGH * expand;
	if (eWIDE > DisplayWidth(theDisp,theScreen)) 
		eWIDE = DisplayWidth(theDisp,theScreen);
	if (eHIGH > DisplayHeight(theDisp,theScreen)) 
		eHIGH = DisplayHeight(theDisp,theScreen);

	eWIDE = width;
	eHIGH = height;

	expImg = gif_Resize(w, eWIDE, eHIGH, *actual_w, *actual_h,
				theImage, NULL);

printf("LOAD w=%x theImg=%x  expImg=%x\n", w, *theImgp, expImg);

	return expImg;
}

/*******************************************/
main(argc, argv)
    int   argc;
    char *argv[];
/*******************************************/
{
    int        i;
    char      *display, *geom, *fname;
    XEvent     event;
	int x, y, width, height;
	Window w;

    cmd = argv[0];
    display = geom = fname = NULL;

    expand = 1;  strip = 0;  nostrip = 0;  mono = 0;

    /*********************Options*********************/

    for (i = 1; i < argc; i++) {
        char *strind;

        if (!strncmp(argv[i],"-g",2)) {		/* geometry */
            i++;
            geom = argv[i];
            continue;
            }

        if (argv[i][0] == '=') {		/* old-style geometry */
            geom = argv[i];
            continue;
            }

        if (!strncmp(argv[i],"-d",2)) {		/* display */
            i++;
            display = argv[i];
            continue;
            }

        strind = index(argv[i], ':');		/* old-style display */
        if(strind != NULL) {
            display = argv[i];
            continue;
            }

        if (!strcmp(argv[i],"-e")) {		/* expand */
            i++;
            expand=atoi(argv[i]);
            continue;
            }

        if (!strcmp(argv[i],"-s")) {		/* strip */
            i++;
            strip=atoi(argv[i]);
            continue;
            }

        if (!strncmp(argv[i],"-n",2)) {		/* nostrip */
            nostrip++;
            continue;
            }

        if (!strncmp(argv[i],"-m",2)) {		/* mono */
            mono++;
            continue;
            }

        if (argv[i][0] != '-') {		/* the file name */
            fname = argv[i];
            continue;
            }

        Syntax(cmd);
    }

    if (fname==NULL) fname="-";
    if (expand<1 || expand>MAXEXPAND) Syntax(cmd);
    if (strip<0 || strip>7) Syntax(cmd);

    /*****************************************************/

    /* Open up the display. */

    if ( (theDisp=XOpenDisplay(display)) == NULL) {
        fprintf(stderr, "%s: Can't open display\n",argv[0]);
        exit(1);
        }

    theScreen = DefaultScreen(theDisp);
    theCmap   = DefaultColormap(theDisp, theScreen);
    rootW     = RootWindow(theDisp,theScreen);
    theGC     = DefaultGC(theDisp,theScreen);
    fcol      = WhitePixel(theDisp,theScreen);
    bcol      = BlackPixel(theDisp,theScreen);
    theVisual = DefaultVisual(theDisp,theScreen);

    dispcells = DisplayCells(theDisp, theScreen);
/*    if (dispcells<=2) 
        FatalError("This program requires a color or greyscale display");
*/

    /****************** Open/Read the File  *****************/

	XParseGeometry(geom,&x,&y,&width,&height);
	x = 0;
	y = 0;
	width = 100;
	height = 100;

	assoc_w[wcount] = gif_CreateMainWindow(x, y, width, height);
	assoc_theImg[wcount] = assoc_expImg[wcount] = NULL;
	assoc_expImg[wcount] = gif_load(fname, w, x, y, width, height,
						&assoc_theW[wcount],
						&assoc_theH[wcount],
						&assoc_theImg[wcount]);
	wcount++;

    /**************** Create/Open X Resources ***************/
    if ((mfinfo = XLoadQueryFont(theDisp,"variable"))==NULL)
       FatalError("couldn't open 'variable' font\n");
    mfont=mfinfo->fid;
    XSetFont(theDisp,theGC,mfont);
    XSetForeground(theDisp,theGC,fcol);
    XSetBackground(theDisp,theGC,bcol);

    /**************** Main loop *****************/
    while (1) {
        XNextEvent(theDisp, &event);
        HandleEvent(&event, x, y, width, height);
        }
}


HandleEvent(event, x, y, width, height)
	XEvent *event;
	int x, y, width, height;
{
    switch (event->type) {
        case Expose: {
            XExposeEvent *exp_event = (XExposeEvent *) event;
	    int idx;

		idx = findAssocIndex(exp_event->window);
		if (idx != -1) {
printf("EXPOSE w=%x theImg=%x  expImg=%x\n",
assoc_w[idx], assoc_theImg[idx],assoc_expImg[idx]);
	                gif_DrawWindow(assoc_w[idx], assoc_expImg[idx],
				exp_event->x,exp_event->y,
           	                exp_event->width, exp_event->height);
		}
            }
            break;

        case KeyPress: {
            XKeyEvent *key_event = (XKeyEvent *) event;
            char buf[128];
            KeySym ks;
            XComposeStatus status;
		Window w;
		char *fname;

            XLookupString(key_event,buf,128,&ks,&status);
            if (buf[0]=='q' || buf[0]=='Q') Quit();

		switch (buf[0]) {
		case '1':
			fname = "/usr/work/gif/mona.gif";
			assoc_w[wcount] = w =
				gif_CreateMainWindow(x, y, width, height);
			assoc_theImg[wcount] = assoc_expImg[wcount] = NULL;
			assoc_expImg[wcount] =
				gif_load(fname, w, x, y, width, height,
						&assoc_theW[wcount],
						&assoc_theH[wcount],
						&assoc_theImg[wcount]);
			++wcount;
		break;
		case '2':
			fname = "/usr/work/gif/avalon_small.gif";
			assoc_w[wcount] = w =
				gif_CreateMainWindow(x, y, width, height);
			assoc_theImg[wcount] = assoc_expImg[wcount] = NULL;
			assoc_expImg[wcount] =
				gif_load(fname, w, x, y, width, height,
						&assoc_theW[wcount],
						&assoc_theH[wcount],
						&assoc_theImg[wcount]);
			++wcount;
		break;
		case '3':
			fname = "/usr/work/gif/zoe";
			assoc_w[wcount] = w =
				gif_CreateMainWindow(x, y, width, height);
			assoc_theImg[wcount] = assoc_expImg[wcount] = NULL;
			assoc_expImg[wcount] =
				gif_load(fname, w, x, y, width, height,
						&assoc_theW[wcount],
						&assoc_theH[wcount],
						&assoc_theImg[wcount]);
			++wcount;
		break;
		}
		buf[0] = '\0';
            }
            break;


        case ConfigureNotify: {
            XConfigureEvent *conf_event = (XConfigureEvent *) event;
	    int idx;
	    int new_width, new_height;

		idx = findAssocIndex(conf_event->window);
	        if (idx != -1)
        		assoc_expImg[idx] = 
			   gif_Resize(assoc_w[idx], 
					conf_event->width, conf_event->height,
					assoc_theW[idx], assoc_theH[idx],
					assoc_theImg[idx],
					assoc_expImg[idx]);
            }
            break;

        case CirculateNotify:
        case MapNotify:
        case DestroyNotify:
        case GravityNotify:
        case ReparentNotify:
        case UnmapNotify:       break;

        default: break;		/* ignore unexpected events */
        }  /* end of switch */
}


/***********************************/
Syntax()
{
    printf("Usage: %s filename [[-geometry] geom] [[-display] display]\n",cmd);
    printf("       [-e 1..%d] [-s 0-7] [-ns] [-mono]\n",MAXEXPAND);
    exit(1);
}


/***********************************/
FatalError (identifier)
       char *identifier;
{
    fprintf(stderr, "%s: %s\n",cmd, identifier);
    exit(-1);
}


/***********************************/
Quit()
{
    exit(0);
}


/***********************************/
gif_DrawWindow(w, img, x, y, width, height)
	Window w;
	XImage *img;
	int x, y, width, height;
{
    XPutImage(theDisp, w, theGC, img,x,y,x,y,width,height);
}


/***********************************/
XImage *gif_Resize(win, w,h, actual_w, actual_h, theImg, expImg)
	Window win;
	int w,h;
	int actual_w, actual_h;
	XImage *theImg;
	XImage *expImg;
{
    int  ix,iy,ex,ey;
    byte *ximag,*ilptr,*ipptr,*elptr,*epptr;
    static char *rstr = "Resizing Image.  Please wait...";

	iWIDE= actual_w;
	iHIGH = actual_h;

/* printf("BEFORE w=%x theImg=%x  expImg=%x\n", win, theImg, expImg); */

    /* warning:  this code'll only run machines where int=32-bits */

    if (w==iWIDE && h==iHIGH) {		/* very special case */
        if (expImg != theImg) {
            if (expImg) XDestroyImage(expImg);
            expImg = theImg;
            eWIDE = iWIDE;  eHIGH = iHIGH;
            }
        }

    else {				/* have to do some work */
        /* if it's a big image, this'll take a while.  mention it */
        if (w*h>(500*500)) {
            XDrawImageString(theDisp,win,theGC,CENTERX(mfinfo,w/2,rstr),
                  CENTERY(mfinfo,h/2),rstr, strlen(rstr));
            XFlush(theDisp);
            }

	/* first, kill the old expImg, if one exists */
	if (expImg && expImg != theImg) {
            free(expImg->data);  expImg->data = NULL;
            XDestroyImage(expImg);
            }

        /* create expImg of the appropriate size */
        
        eWIDE = w;  eHIGH = h;
        ximag = (byte *) malloc(w*h);
	if (dispcells <= 2) {
	  expImg = XCreateImage(theDisp,theVisual,1,ZPixmap,0,ximag,
				eWIDE,eHIGH,8,eWIDE);
	} else {
	  expImg = XCreateImage(theDisp,theVisual,8,ZPixmap,0,ximag,
				eWIDE,eHIGH,8,eWIDE);
	}

        if (!ximag || !expImg) {
            fprintf(stderr,"ERROR: unable to create a %dx%d image\n",w,h);
            exit(0);
            }

        elptr = epptr = (byte *) expImg->data;

        for (ey=0;  ey<eHIGH;  ey++, elptr+=eWIDE) {
	  extern byte *Image;/*XXX ICK*/

	  iy = (iHIGH * ey) / eHIGH;
/*	  printf("iy=%d   :", iy);*/
	  epptr = elptr;
/*	  ilptr = (byte *) theImg->data + (iy * iWIDE);*/
	  ilptr = (byte *) Image + (iy * iWIDE);
	  for (ex=0;  ex<eWIDE;  ex++,epptr++) {
	    ix = (iWIDE * ex) / eWIDE;
	    ipptr = ilptr + ix;
/*	    printf("%d ", *ipptr);*/
	    *epptr = *ipptr;
	    if (dispcells <= 2) {
	      XPutPixel(expImg, ex, ey, (unsigned long)*ipptr);
	    }
	  }
/*	  printf("\n");*/
	}
      }

/* printf("AFTER ..... expImg=%x\n", expImg); */

    return expImg;
}
                

/***********************************/
Window gif_CreateMainWindow(x, y, width, height)
	int x, y, width, height;
{
    XSetWindowAttributes xswa;
    unsigned int xswamask;
    XSizeHints hints;
	int i = 1;

    if (i&XValue && i&XNegative) 
        x = XDisplayWidth(theDisp,theScreen)-eWIDE-abs(x);
    if (i&YValue && i&YNegative) 
        y = XDisplayHeight(theDisp,theScreen)-eHIGH-abs(y);

    hints.x=x;             hints.y=y;
    hints.width  = eWIDE;  hints.height = eHIGH;
    hints.max_width = DisplayWidth(theDisp,theScreen);
    hints.max_height = DisplayHeight(theDisp,theScreen);
    hints.flags |= PMaxSize;

    xswa.background_pixel = bcol;
    xswa.border_pixel     = fcol;
    xswamask = CWBackPixel | CWBorderPixel;

    mainW = XCreateWindow(theDisp,rootW,x,y,width,height,2,0,CopyFromParent,
                          CopyFromParent, xswamask, &xswa);

    if (!mainW) FatalError("Can't open main window");
/*
    XSetStandardProperties(theDisp,mainW,"xgif","xgif",None,
                            argv,argc,&hints);
*/
    XSelectInput(theDisp, mainW, ExposureMask | KeyPressMask 
                               | StructureNotifyMask);
    XMapWindow(theDisp, mainW);

	return mainW;
}

