/*		Our Static DTD for HTML
**		-----------------------
**
**	 6 Nov 93	MD	Increased size of img_attr array to make space
**  				for terminator.
**	20 Dec 93	PYW	Added HTML+ stuff; valid-subelements DTD info.
*/

/* Implements:
*/

#include "HTMLDTD.h"

/* 	Entity Names
**	------------
**
**	This table must be matched exactly with ALL the translation tables
*/
static CONST char* entities[] = {
  "AElig",	/* capital AE diphthong (ligature) */ 
  "Aacute",	/* capital A, acute accent */ 
  "Acirc",	/* capital A, circumflex accent */ 
  "Agrave",	/* capital A, grave accent */ 
  "Aring",	/* capital A, ring */ 
  "Atilde",	/* capital A, tilde */ 
  "Auml",	/* capital A, dieresis or umlaut mark */ 
  "Ccedil",	/* capital C, cedilla */ 
  "ETH",	/* capital Eth, Icelandic */ 
  "Eacute",	/* capital E, acute accent */ 
  "Ecirc",	/* capital E, circumflex accent */ 
  "Egrave",	/* capital E, grave accent */ 
  "Euml",	/* capital E, dieresis or umlaut mark */ 
  "Iacute",	/* capital I, acute accent */ 
  "Icirc",	/* capital I, circumflex accent */ 
  "Igrave",	/* capital I, grave accent */ 
  "Iuml",	/* capital I, dieresis or umlaut mark */ 
  "Ntilde",	/* capital N, tilde */ 
  "Oacute",	/* capital O, acute accent */ 
  "Ocirc",	/* capital O, circumflex accent */ 
  "Ograve",	/* capital O, grave accent */ 
  "Oslash",	/* capital O, slash */ 
  "Otilde",	/* capital O, tilde */ 
  "Ouml",	/* capital O, dieresis or umlaut mark */ 
  "THORN",	/* capital THORN, Icelandic */ 
  "Uacute",	/* capital U, acute accent */ 
  "Ucirc",	/* capital U, circumflex accent */ 
  "Ugrave",	/* capital U, grave accent */ 
  "Uuml",	/* capital U, dieresis or umlaut mark */ 
  "Yacute",	/* capital Y, acute accent */ 
  "aacute",	/* small a, acute accent */ 
  "acirc",	/* small a, circumflex accent */ 
  "aelig",	/* small ae diphthong (ligature) */ 
  "agrave",	/* small a, grave accent */ 
  "amp",	/* ampersand */ 
  "aring",	/* small a, ring */ 
  "atilde",	/* small a, tilde */ 
  "auml",	/* small a, dieresis or umlaut mark */ 
  "ccedil",	/* small c, cedilla */ 
  "eacute",	/* small e, acute accent */ 
  "ecirc",	/* small e, circumflex accent */ 
  "egrave",	/* small e, grave accent */ 
  "emsp",       /* emsp, em space - not collapsed */
  "ensp",       /* ensp, en space - not collapsed */
  "eth",	/* small eth, Icelandic */ 
  "euml",	/* small e, dieresis or umlaut mark */ 
  "gt",		/* greater than */ 
  "iacute",	/* small i, acute accent */ 
  "icirc",	/* small i, circumflex accent */ 
  "igrave",	/* small i, grave accent */ 
  "iuml",	/* small i, dieresis or umlaut mark */ 
  "ldquo",	/* left double quote */
  "lt",		/* less than */ 
  "nbsp",	/* nbsp, non breaking space */
  "ntilde",	/* small n, tilde */ 
  "oacute",	/* small o, acute accent */ 
  "ocirc",	/* small o, circumflex accent */ 
  "ograve",	/* small o, grave accent */ 
  "oslash",	/* small o, slash */ 
  "otilde",	/* small o, tilde */ 
  "ouml",	/* small o, dieresis or umlaut mark */ 
  "quote",	/* quote */
  "rdquo",	/* right double quote */
  "szlig",	/* small sharp s, German (sz ligature) */ 
  "thorn",	/* small thorn, Icelandic */ 
  "uacute",	/* small u, acute accent */ 
  "ucirc",	/* small u, circumflex accent */ 
  "ugrave",	/* small u, grave accent */ 
  "uuml",	/* small u, dieresis or umlaut mark */ 
  "yacute",	/* small y, acute accent */ 
  "yuml",	/* small y, dieresis or umlaut mark */ 
};

#define HTML_ENTITIES 65


/*		Attribute Lists
**		---------------
**
**	Lists must be in alphatbetical order by attribute name
**	The tag elements contain the number of attributes
*/

static attr no_attr[] = 
	{{ 0 }};

static attr a_attr[] = {				/* Should be ID */
	{ "HREF" },
	{ "NAME" },
	{ "REL" },
	{ "REV" },
	{ "TITLE" },
	{ "TYPE" },
	{ "URN" },
	{0}
};

static attr address_attr[] = {
	{ "BORDER" },
	{ "FOLD" },
	{ "LABEL" },
	{0}
};

static attr author_attr[] = {
	{ "HREF" },
	{0}
};

static attr emph_attr[] = {	
	{ "B" },
	{ "I" },
	{ "INDEX" },
	{ "SUB" },
	{ "SUP" },
	{ "TT" },
	{ "TYPE" },
	{0}
};

static attr entry_attr[] = {	
	{ "HREF" },
	{ "NAME" },
	{ "TYPE" },
	{0}
};

static attr figa_attr[] = {	
	{ "AREA" },
	{ "HREF" },
	{0}
};

static attr figure_attr[] = {	
	{ "FOLD" },
	{ "HEIGHT" },
	{ "ISMAP" },
	{ "LABEL" },
	{ "MAXWIDTH" },
	{ "MAYDELAY" },
	{ "MINWIDTH" },
	{ "SRC" },
	{ "TYPE" },
	{ "WIDTH" },
	{0}
};

static attr form_attr[] = {	
	{ "ACTION" },
	{ "ENVVAR" },
	{ "FOLD" },
	{ "LABEL" },
	{ "MAXWIDTH" },
	{ "METHOD" },
	{ "MINWIDTH" },
	{ "NAME" },
	{0}
};

static attr header_attr[] = {	
	{ "ID" },
	{ "NAME" },
	{0}
};

static attr hpane_attr[] = {	
	{ "CENTER" },
	{ "ID" },
	{ "LEFT" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{ "RIGHT" },
	{ "STYLE" },
	{0}
};

static attr img_attr[] = {	
	{ "ALT" },
	{ "BGAP" },
	{ "HREF" },
	{ "ISMAP" },
	{ "LGAP" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{ "RGAP" },
	{ "SRC" },
	{ "TGAP" },
	{0}
};

static attr input_attr[] = {	
	{ "CHARS" },
	{ "CHECKED" },
	{ "DISABLED" },
	{ "ENVVAR" },
	{ "MAXCHARS" },
	{ "MAXWIDTH" },
	{ "MINCHARS" },
	{ "MINWIDTH" },
	{ "NAME" },
	{ "SIZE" },
	{ "TYPE" },
	{ "VALUE" },
	{0}
};

static attr insert_attr[] = {	
	{ "ASIS" },
	{ "BORDER" },
	{ "HREF" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{0}
};

static attr l_attr[] = {	
	{ "ALIGN" },
	{ "ID" },
	{ "INDEX" },
	{ "LANG" },
	{0}
};

static attr link_attr[] = {	
	{ "ARG" },
	{ "HREF" },
	{ "ID" },
	{ "REL" },
	{ "REV" },
	{0}
};

static attr menu_attr[] = {
	{ "COMPACT" },
	{ "FOLD" },
	{ "LABEL" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{0}
};

static attr mh_attr[] = {	
	{ "HIDDEN" },
	{0}
};

static attr nextid_attr[] = {
	{ "N" },
	{ 0 }	/* Terminate list */
};

static attr ol_attr[] = {
	{ "COMPACT" },
	{ "FOLD" },
	{ "LABEL" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{ "START" },
	{0}
};

static attr option_attr[] = {	
	{ "CHECKED" },
	{ "DISABLED" },
	{ "NAME" },
	{ "VALUE" },
	{0}
};

static attr select_attr[] = {	
	{ "CHECKED" },
	{ "DISABLED" },
	{ "ENVVAR" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{ "NAME" },
	{ "SIZE" },
	{ "TYPE" },
	{ "VALUE" },
	{0}
};

static attr p_attr[] = {	
	{ "ALIGN" },
	{ "FOLD" },
	{ "ID" },
	{ "INDEX" },
	{ "LABEL" },
	{ "LANG" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{ "STYLE" },
	{0}
};

static attr pre_attr[] = {	
	{ "WIDTH" },
	{0}
};

static attr section_attr[] = {	
	{ "BORDER" },
	{ "FOLD" },
	{ "HREF" },
	{ "LABEL" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{ "NAME" },
	{0}
};

static attr security_attr[] = {	
	{ "LEVEL" },
	{0}
};

static attr table_attr[] = {	
	{ "BORDER" },
	{ "COMPACT" },
	{ "FOLD" },
	{ "ID" },
	{ "INDEX" },
	{ "LABEL" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{0}
};

static attr tcap_attr[] = {	
	{ "ALIGN" },
	{ "TOP" },
	{0}
};

static attr td_attr[] = {	
	{ "ALIGN" },
	{ "COLSPAN" },
	{ "ROWSPAN" },
	{0}
};

static attr textarea_attr[] = {	
	{ "COL" },
	{ "NAME" },
	{ "ROW" },
	{0}
};

static attr th_attr[] = {	
	{ "ALIGN" },
	{ "COLSPAN" },
	{ "ROWSPAN" },
	{0}
};

static attr ul_attr[] = {
	{ "COMPACT" },
	{ "FOLD" },
	{ "LABEL" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{0}
};

static attr vobjf_attr[] = {	
	{ "ARG" },
	{ "HREF" },
	{ "MAXWIDTH" },
	{ "MINWIDTH" },
	{ "NAME" },
	{0}
};

/* messy */
static attr shared_list_attr[] = {
	{ "COMPACT" },
	{ 0 }	/* Terminate list */
};

#define dd_dt_validSubElementsCount 77
static HTMLElement dd_dt_validSubElements[] = {
        HTML_A,
        HTML_ADDRESS,
        HTML_AUTHOR,
        HTML_B,
        HTML_BLOCKQUOTE,
        HTML_BODY,
        HTML_BOLD,
        HTML_BR,
        HTML_CAUTION,
        HTML_CITE,
        HTML_CMD,
        HTML_CODE,
        HTML_COMMENT,
        HTML_DFN,
        HTML_DIR,
        HTML_DL,
        HTML_DLC,
        HTML_EM,
        HTML_EMPH,
        HTML_ENTRY,
        HTML_EXAMPLE,
        HTML_FIGURE,
	HTML_FOOTNOTE,
        HTML_FORM,
        HTML_H1,
        HTML_H2,
        HTML_H3,
        HTML_H4,
        HTML_H5,
        HTML_H6,
        HTML_H7,
        HTML_HEAD,
        HTML_HPANE,
        HTML_HR,
        HTML_HTML,
        HTML_I,
        HTML_IMAGE,
        HTML_IMG,
        HTML_INPUT,
	HTML_INSERT,
        HTML_ISINDEX,
        HTML_ITALIC,
        HTML_KBD,
        HTML_KEY,
        HTML_L,
        HTML_LINK,
        HTML_LISTING,
        HTML_LIT,
        HTML_MENU,
        HTML_NEXTID,
        HTML_OL,
        HTML_P,
        HTML_PLAIN,
        HTML_PLAINTEXT,
        HTML_PRE,
        HTML_QUOTE,
        HTML_SAMP,
        HTML_SCREEN,
        HTML_SECTION,
        HTML_SELECT,
        HTML_STRONG,
        HTML_SUB,
        HTML_SUP,
        HTML_TABLE,
        HTML_TEXTAREA,
        HTML_TITLE,
        HTML_TT,
        HTML_U,
        HTML_UL,
        HTML_UNDER,
        HTML_VAR,
        HTML_VINSERT,
        HTML_VOBJF,
        HTML_VR,
        HTML_VSCRIPT,
        HTML_VVIEW,
        HTML_XMP,
};

#define l_validSubElementsCount 3
static HTMLElement l_validSubElements[] = {
	HTML_A,
	HTML_IMAGE,
	HTML_IMG,
};

/* Oh hell. Important thing is to omit LI. Plus other misc sub-tags */
#define li_validSubElementsCount 77
static HTMLElement li_validSubElements[] = {
        HTML_A,
        HTML_ADDRESS,
        HTML_AUTHOR,
        HTML_B,
        HTML_BLOCKQUOTE,
        HTML_BODY,
        HTML_BOLD,
        HTML_BR,
        HTML_CAUTION,
        HTML_CITE,
        HTML_CMD,
        HTML_CODE,
        HTML_COMMENT,
        HTML_DFN,
        HTML_DIR,
        HTML_DL,
        HTML_DLC,
        HTML_EM,
        HTML_EMPH,
        HTML_ENTRY,
        HTML_EXAMPLE,
        HTML_FIGURE,
	HTML_FOOTNOTE,
        HTML_FORM,
        HTML_H1,
        HTML_H2,
        HTML_H3,
        HTML_H4,
        HTML_H5,
        HTML_H6,
        HTML_H7,
        HTML_HEAD,
        HTML_HPANE,
        HTML_HR,
        HTML_HTML,
        HTML_I,
        HTML_IMAGE,
        HTML_IMG,
        HTML_INPUT,
	HTML_INSERT,
        HTML_ISINDEX,
        HTML_ITALIC,
        HTML_KBD,
        HTML_KEY,
        HTML_L,
        HTML_LINK,
        HTML_LISTING,
        HTML_LIT,
        HTML_MENU,
        HTML_NEXTID,
        HTML_OL,
        HTML_P,
        HTML_PLAIN,
        HTML_PLAINTEXT,
        HTML_PRE,
        HTML_QUOTE,
        HTML_SAMP,
        HTML_SCREEN,
        HTML_SECTION,
        HTML_SELECT,
        HTML_STRONG,
        HTML_SUB,
        HTML_SUP,
        HTML_TABLE,
        HTML_TEXTAREA,
        HTML_TITLE,
        HTML_TT,
        HTML_U,
        HTML_UL,
        HTML_UNDER,
        HTML_VAR,
        HTML_VINSERT,
        HTML_VOBJF,
        HTML_VR,
        HTML_VSCRIPT,
        HTML_VVIEW,
        HTML_XMP,
};

#define ol_ul_validSubElementsCount 30
static HTMLElement ol_ul_validSubElements[] = {
        HTML_A,
        HTML_B,
        HTML_BOLD,
        HTML_BR,
        HTML_CITE,
        HTML_CMD,
        HTML_CODE,
        HTML_DFN,
        HTML_EM,
        HTML_EMPH,
	HTML_FOOTNOTE,
        HTML_I,
        HTML_IMAGE,
        HTML_IMG,
        HTML_ITALIC,
        HTML_KBD,
        HTML_KEY,
        HTML_L,
        HTML_LI,
        HTML_LINK,
        HTML_P,
        HTML_SAMP,
        HTML_STRONG,
        HTML_SUB,
        HTML_SUP,
        HTML_TT,
        HTML_U,
        HTML_UNDER,
        HTML_VAR,
        HTML_VINSERT,
};

#define p_validSubElementsCount 29
static HTMLElement p_validSubElements[] = {
        HTML_A,
        HTML_B,
        HTML_BOLD,
        HTML_CITE,
        HTML_CMD,
        HTML_CODE,
        HTML_DFN,
        HTML_EM,
        HTML_EMPH,
	HTML_FOOTNOTE,
        HTML_I,
        HTML_IMAGE,
        HTML_IMG,
        HTML_INPUT,
        HTML_ITALIC,
        HTML_KBD,
        HTML_KEY,
        HTML_L,
        HTML_LINK,
        HTML_SAMP,
        HTML_STRONG,
        HTML_SUB,
        HTML_SUP,
        HTML_TT,
        HTML_U,
        HTML_UNDER,
        HTML_VAR,
        HTML_VINSERT,
};


#define option_validSubElementsCount 28
static HTMLElement option_validSubElements[] = {
        HTML_A,
        HTML_B,
        HTML_BOLD,
        HTML_BR,
        HTML_CITE,
        HTML_CMD,
        HTML_CODE,
        HTML_DFN,
        HTML_EM,
        HTML_EMPH,
	HTML_FOOTNOTE,
        HTML_I,
        HTML_IMAGE,
        HTML_IMG,
        HTML_ITALIC,
        HTML_KBD,
        HTML_KEY,
        HTML_L,
        HTML_LINK,
        HTML_SAMP,
        HTML_STRONG,
        HTML_SUB,
        HTML_SUP,
        HTML_TT,
        HTML_U,
        HTML_UNDER,
        HTML_VAR,
        HTML_VINSERT,
};

#define th_validSubElementsCount 33
static HTMLElement th_validSubElements[] = {
        HTML_A,
        HTML_B,
        HTML_BOLD,
        HTML_BR,
        HTML_CITE,
        HTML_CMD,
        HTML_CODE,
        HTML_DFN,
        HTML_DIR,
        HTML_DL,
        HTML_EM,
        HTML_EMPH,
	HTML_FOOTNOTE,
        HTML_I,
        HTML_IMAGE,
        HTML_IMG,
        HTML_ITALIC,
        HTML_KBD,
        HTML_KEY,
        HTML_L,
        HTML_LINK,
        HTML_MENU,
        HTML_OL, 
        HTML_SAMP,
        HTML_STRONG,
        HTML_SUB,
        HTML_SUP,
        HTML_TT,
        HTML_U,
        HTML_UL,
        HTML_UNDER,
        HTML_VAR,
        HTML_VINSERT,
};

#define loose_validSubElementsCount 78
static HTMLElement loose_validSubElements[] = {
        HTML_A,
        HTML_ADDRESS,
        HTML_AUTHOR,
        HTML_B,
        HTML_BLOCKQUOTE,
        HTML_BODY,
        HTML_BOLD,
        HTML_BR,
        HTML_CAUTION,
        HTML_CITE,
        HTML_CMD,
        HTML_CODE,
        HTML_COMMENT,
        HTML_DFN,
        HTML_DIR,
        HTML_DL,
        HTML_DLC,
        HTML_EM,
        HTML_EMPH,
        HTML_ENTRY,
        HTML_EXAMPLE,
        HTML_FIGURE,
	HTML_FOOTNOTE,
        HTML_FORM,
        HTML_H1,
        HTML_H2,
        HTML_H3,
        HTML_H4,
        HTML_H5,
        HTML_H6,
        HTML_H7,
        HTML_HEAD,
        HTML_HPANE,
        HTML_HR,
        HTML_HTML,
        HTML_I,
        HTML_IMAGE,
        HTML_IMG,
        HTML_INPUT,
	HTML_INSERT,
        HTML_ISINDEX,
        HTML_ITALIC,
        HTML_KBD,
        HTML_KEY,
        HTML_L,
        HTML_LINK,
        HTML_LISTING,
        HTML_LIT,
        HTML_MENU,
        HTML_NEXTID,
        HTML_OL,
        HTML_P,
        HTML_PLAIN,
        HTML_PLAINTEXT,
        HTML_PRE,
        HTML_QUOTE,
        HTML_SAMP,
        HTML_SCREEN,
        HTML_SECTION,
        HTML_SELECT,
        HTML_SECURITY,
        HTML_STRONG,
        HTML_SUB,
        HTML_SUP,
        HTML_TABLE,
        HTML_TEXTAREA,
        HTML_TITLE,
        HTML_TT,
        HTML_U,
        HTML_UL,
        HTML_UNDER,
        HTML_VAR,
        HTML_VINSERT,
        HTML_VOBJF,
        HTML_VR,
        HTML_VSCRIPT,
        HTML_VVIEW,
        HTML_XMP
};

/*	Elements
*/
/*    Name, 	Attributes, 		content
*/
static HTTag tags[] = {
    { "A"       , a_attr,       HTML_A_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "ADDRESS" , address_attr, HTML_ADDRESS_ATTRIBUTES,SGML_MIXED, 0, 0},
    { "AUTHOR"  , author_attr,  HTML_AUTHOR_ATTRIBUTES, SGML_EMPTY, 0, 0},
    { "B"       , no_attr,      0,			SGML_MIXED, 0, 0},
    { "BLOCKQUOTE", no_attr,    0,			SGML_MIXED, 0, 0},
    { "BODY"    , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "BOLD"    , no_attr,      0,             		SGML_MIXED, 0, 0},
/*  { "BR"      , no_attr,      0,             		SGML_EMPTY, 0, 0}, */
/* kudge */
    { "BR"      , p_attr,       HTML_P_ATTRIBUTES,	SGML_MIXED,
						p_validSubElements,
						p_validSubElementsCount},
    { "CAUTION" , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "CITE"    , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "CMD"     , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "CODE"    , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "COMMENT" , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "DD"      , no_attr,      0,             		SGML_MIXED,
						dd_dt_validSubElements,
						dd_dt_validSubElementsCount},
    { "DFN"     , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "DIR"     , shared_list_attr,1,          		SGML_MIXED, 0, 0},/*?*/
    { "DL"      , shared_list_attr,1,          		SGML_MIXED, 0, 0},
    { "DLC"     , shared_list_attr,1,          		SGML_MIXED, 0, 0},
    { "DT"      , no_attr,      0,             		SGML_MIXED,
						dd_dt_validSubElements,
						dd_dt_validSubElementsCount},
    { "EM"      , emph_attr,    HTML_EMPH_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "EMPH"    , emph_attr,    HTML_EMPH_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "ENTRY"   , entry_attr,   HTML_ENTRY_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "EXAMPLE" , no_attr,      0,             		SGML_LITTERAL, 0, 0},
    { "FIGA"    , figa_attr,    HTML_FIGA_ATTRIBUTES,	SGML_EMPTY, 0, 0},
    { "FIGCAP"  , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "FIGDATA" , no_attr,      0,             		SGML_LITTERAL, 0, 0},
    { "FIG" 	, figure_attr,  HTML_FIGURE_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "FIGURE"  , figure_attr,  HTML_FIGURE_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "FOOTNOTE", no_attr,      0,			SGML_MIXED, 0, 0},
    { "FORM"    , form_attr,    HTML_FORM_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "GROUP"   , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "H1"      , header_attr,  HTML_HEADER_ATTRIBUTES, SGML_MIXED, 
						p_validSubElements,
						p_validSubElementsCount},
    { "H2"      , header_attr,  HTML_HEADER_ATTRIBUTES, SGML_MIXED,
						p_validSubElements,
						p_validSubElementsCount},
    { "H3"      , header_attr,  HTML_HEADER_ATTRIBUTES, SGML_MIXED,
						p_validSubElements,
						p_validSubElementsCount},
    { "H4"      , header_attr,  HTML_HEADER_ATTRIBUTES, SGML_MIXED,
						p_validSubElements,
						p_validSubElementsCount},
    { "H5"      , header_attr,  HTML_HEADER_ATTRIBUTES, SGML_MIXED,
						p_validSubElements,
						p_validSubElementsCount},
    { "H6"      , header_attr,  HTML_HEADER_ATTRIBUTES, SGML_MIXED,
						p_validSubElements,
						p_validSubElementsCount},
    { "H7"      , header_attr,  HTML_HEADER_ATTRIBUTES, SGML_MIXED,
						p_validSubElements,
						p_validSubElementsCount},
    { "HEAD"    , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "HPANE"   , hpane_attr,   HTML_HPANE_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "HR"      , no_attr,      0,             		SGML_EMPTY, 0, 0},
    { "HTML"    , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "I"       , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "IMAGE"   , img_attr,     HTML_IMG_ATTRIBUTES,	SGML_EMPTY, 0, 0},
    { "IMG"     , img_attr,     HTML_IMG_ATTRIBUTES,	SGML_EMPTY, 0, 0},
    { "INPUT"   , input_attr,	HTML_INPUT_ATTRIBUTES,	SGML_EMPTY, 0, 0},
    { "INSERT"  , insert_attr,	HTML_INSERT_ATTRIBUTES, SGML_EMPTY, 0, 0},
    { "ISINDEX" , no_attr,      0,	             	SGML_EMPTY, 0, 0},
    { "ITALIC"  , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "KBD"     , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "KEY"     , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "L"       , l_attr,	HTML_L_ATTRIBUTES,	SGML_MIXED, 
						l_validSubElements,
						l_validSubElementsCount},
    { "LI"      , shared_list_attr,1,          		SGML_MIXED,
						li_validSubElements,
						li_validSubElementsCount},
    { "LINK"    , link_attr,    HTML_LINK_ATTRIBUTES,	SGML_EMPTY, 0, 0},
    { "LISTING" , no_attr,      0,			SGML_MIXED, 0, 0},
    { "LIT"     , no_attr,      0,             		SGML_LITTERAL, 0, 0},
    { "MENU"    , menu_attr,    HTML_MENU_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "MH"      , mh_attr,      HTML_MH_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "NEXTID"  , nextid_attr,  1,             		SGML_EMPTY, 0, 0},
    { "OL"      , ol_attr,      HTML_OL_ATTRIBUTES,	SGML_MIXED, 
						ol_ul_validSubElements,
						ol_ul_validSubElementsCount},
    { "OLC"      , ol_attr,      HTML_OL_ATTRIBUTES,	SGML_MIXED, 
						ol_ul_validSubElements,
						ol_ul_validSubElementsCount},
    { "OPTION"  , option_attr,  HTML_OPTION_ATTRIBUTES,	SGML_MIXED,
						option_validSubElements,
    						option_validSubElementsCount},
    { "P"       , p_attr,       HTML_P_ATTRIBUTES,	SGML_MIXED,
						p_validSubElements,
						p_validSubElementsCount},
    { "PLAIN"   , no_attr,      0,             		SGML_LITTERAL, 0, 0},
    { "PLAINTEXT", no_attr,     0,             		SGML_LITTERAL, 0, 0},
    { "PRE"     , pre_attr,     HTML_PRE_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "QUOTE"   , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "SAMP"    , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "SCREEN"  , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "SECTION" , section_attr, HTML_SECTION_ATTRIBUTES,SGML_MIXED, 0, 0},
    { "SECURITY", security_attr,HTML_SECURITY_ATTRIBUTES,SGML_EMPTY, 0, 0},
    { "SELECT"  , select_attr,  HTML_SELECT_ATTRIBUTES, SGML_MIXED, 0, 0},
    { "STRONG"  , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "SUB"     , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "SUP"     , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "TABLE"   , table_attr,   HTML_TABLE_ATTRIBUTES,	SGML_MIXED, 0, 0},
    { "TB"      , no_attr,      0,             		SGML_EMPTY, 0, 0},
    { "TCAP"    , tcap_attr,    HTML_TCAP_ATTRIBUTES,	SGML_CDATA, 0, 0},
    { "TD"      , td_attr,      HTML_TD_ATTRIBUTES,	SGML_MIXED,
						loose_validSubElements,
						loose_validSubElementsCount},
    { "TEXTAREA", textarea_attr, HTML_TEXTAREA_ATTRIBUTES, SGML_LITTERAL,0,0},
    { "TH"      , th_attr,      HTML_TH_ATTRIBUTES,	SGML_CDATA, 
						th_validSubElements,
						th_validSubElementsCount},
    { "TITLE"   , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "TR"      , no_attr,      0,             		SGML_EMPTY, 0, 0},
    { "TT"      , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "U"       , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "UL"      , ul_attr,      HTML_UL_ATTRIBUTES,	SGML_MIXED,
						ol_ul_validSubElements,
						ol_ul_validSubElementsCount},
    { "ULC"      , ul_attr,      HTML_UL_ATTRIBUTES,	SGML_MIXED,
						ol_ul_validSubElements,
						ol_ul_validSubElementsCount},
    { "UNDER"   , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "VAR"     , no_attr,      0,             		SGML_MIXED, 0, 0},
    { "VINSERT" , no_attr,      0,             		SGML_LITTERAL, 0, 0},
    { "VOBJF"   , vobjf_attr,   HTML_VOBJF_ATTRIBUTES,	SGML_EMPTY, 0, 0},
    { "VR"      , no_attr,      0,             		SGML_EMPTY, 0, 0},
    { "VSCRIPT" , no_attr,      0,             		SGML_LITTERAL, 0, 0},
    { "VVIEW"   , no_attr,      0,             		SGML_LITTERAL, 0, 0},
    { "XMP"     , no_attr,      0,             		SGML_LITTERAL, 0, 0},
    {0}
};
#define HTML_TAGS 95


PUBLIC CONST SGML_dtd HTML_dtd = {
	tags,
	HTML_ELEMENTS,
	entities,
	sizeof(entities)/sizeof(char**)
};

/*	Utility Routine: useful for people building HTML objects */

/*	Start anchor element
**	--------------------
**
**	It is kinda convenient to have a particulr routine for
**	starting an anchor element, as everything else for HTML is
**	simple anyway.
*/
struct _HTStructured {
    HTStructuredClass * isa;
	/* ... */
};

PUBLIC void HTStartAnchor ARGS3(HTStructured *, obj,
		CONST char *,  name,
		CONST char *,  href)
{
    BOOL		present[HTML_A_ATTRIBUTES];
    CONST char*		value[HTML_A_ATTRIBUTES];
    
    {
    	int i;
    	for(i=0; i<HTML_A_ATTRIBUTES; i++)
	    present[i] = NO;
    }
    if (name) {
    	present[HTML_A_NAME] = YES;
	value[HTML_A_NAME] = name;
    }
    if (href) {
        present[HTML_A_HREF] = YES;
        value[HTML_A_HREF] = href;
    }
    
    (*obj->isa->start_element)(obj, HTML_A , present, value,
    					&HTML_dtd.tags[HTML_A]/*PYW*/);

}

PUBLIC void HTNextID ARGS2(HTStructured *, obj,
		int,	next_one)
{
    BOOL		present[HTML_NEXTID_ATTRIBUTES];
    CONST char*		value[HTML_NEXTID_ATTRIBUTES];
    char string[10];
    
    sprintf(string, "z%i", next_one);
    {
    	int i;
    	for(i=0; i<HTML_NEXTID_ATTRIBUTES; i++)
	    present[i] = NO;
    }
    present[HTML_NEXTID_N] = YES;
    value[HTML_NEXTID_N] = string;
    
    (*obj->isa->start_element)(obj, HTML_NEXTID , present, value,
    					&HTML_dtd.tags[HTML_NEXTID]/*PYW*/);
}

